/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**@file
 * config.
 */

#ifndef OBS_CONFIG_H_
#define OBS_CONFIG_H_

#include <stddef.h>
#include <obs/obs_host_config.h>

/* ----------------------------------------------------------------------
 * flags
 */

/*
 * language check
 */

/* ISO C */
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
# define OBS_ISOC99_			1
# ifndef OBS_HAS_STRUCT_ELEM_INIT_
#  define OBS_HAS_STRUCT_ELEM_INIT_	1
# endif
# ifndef OBS_HAS_ARRAY_ELEM_INIT_
#  define OBS_HAS_ARRAY_ELEM_INIT_	1
# endif
#endif


/* C++ */
#ifdef __cplusplus
# define OBS_CPLUSPLUS_		1
#endif

/* Doxygen */
#ifdef OBS_DOXYGEN_
/**@internal whether structure element initializer can be used. */
# define OBS_HAS_STRUCT_ELEM_INIT_	1
/**@internal whether use array element initializer can be used. */
# define OBS_HAS_ARRAY_ELEM_INIT_	1
#endif /* Doxygen */


/*
 * compiler specific
 */

/* GCC */
#ifdef __GNUC__
#include <obs/obs_gcc_config.h>
#endif /* __GNUC__ */


/* ----------------------------------------------------------------------
 * flag defaults
 */

#ifndef OBS_HAS_STRUCT_ELEM_INIT_
# define OBS_HAS_STRUCT_ELEM_INIT_	0
#endif

#ifndef OBS_HAS_ARRAY_ELEM_INIT_
# define OBS_HAS_ARRAY_ELEM_INIT_	0
#endif

#ifndef OBS_CPLUSPLUS_
/**@internal whether C++ */
# define OBS_CPLUSPLUS_			0
#endif

#ifndef OBS_ISOC99_
/** @internal whether ISO C 99. */
# define OBS_ISOC99_	0
#endif

/*
 * features
 */

/* Doxygen */
#ifdef OBS_DOXYGEN_
/**@internal default specifier for inline function. */
# define OBS_INLINE_	inline
#endif

/* lint */
#ifdef lint
#define OBS_INLINE_	static inline
# define OBS_DEF_INLINE_FUNC_AU__(ret, name, args)			      \
OBS_INLINE_ ret /*ARGSUSED*/ name args
#endif

/*
 * function declaration.
 */
#if OBS_CPLUSPLUS_
# define OBS_BEGIN_DECL__	extern "C" {
# define OBS_END_DECL__		}
#endif

#if OBS_HAS_STRUCT_ELEM_INIT_
/**@internal initializer for a structure element. */
# define OBS_STRUCT_INIT__(name, value)	.name = (value)
#endif

#if OBS_HAS_ARRAY_ELEM_INIT_
/**@internal initializer for an array element. */
# define OBS_ARRAY_INIT__(suf, value)	[(suf)] = (value)
#endif


/*
 * features defaults
 */

#ifndef OBS_INLINE_
# define OBS_INLINE_	static
#endif

#ifndef OBS_DEF_INLINE_FUNC__
# ifndef OBS_NO_INLINE
/**@internal define an inline function. */
#  define OBS_DEF_INLINE_FUNC__(r, n, a)		OBS_INLINE_ r n a
# else
#  define OBS_DEF_INLINE_FUNC__(r, n, a)		static r n a
# endif
#endif

#ifndef OBS_DEF_INLINE_FUNC_AU__
/**@internal define an inline function which has unused argument(s). */
# define OBS_DEF_INLINE_FUNC_AU__(r, n, a)	OBS_DEF_INLINE_FUNC__(r,n,a)
#endif

#ifndef OBS_BEGIN_DECL__
/**@internal begin of function prototype. */
# define OBS_BEGIN_DECL__
#endif

#ifndef OBS_END_DECL__
/**@internal end of function prototype. */
# define OBS_END_DECL__
#endif

#ifndef OBS_STRUCT_INIT__
# define OBS_STRUCT_INIT__(name, value)	(value)
#endif

#ifndef OBS_ARRAY_INIT__
# define OBS_ARRAY_INIT__(suf, value)	(value)
#endif

#ifndef OBS_UINTPTR_TYPE_
# error OBS_UINTPTR_TYPE_ is not defined.
#endif

#ifndef OBS_INTPTR_TYPE_
# error OBS_INTPTR_TYPE_ is not defined.
#endif

#ifndef OBS_INT8_TYPE_
/**@internal signed 8bit integer type. */
# define OBS_INT8_TYPE_	signed char
#endif

#ifndef OBS_UINT8_TYPE_
/**@internal unsigned 8bit integer type. */
# define OBS_UINT8_TYPE_	unsigned char
#endif

#ifndef OBS_INT16_TYPE_
# error OBS_INT16_TYPE_ is not defined.
#endif

#ifndef OBS_UINT16_TYPE_
# error OBS_UINT16_TYPE_ is not defined.
#endif

#ifndef OBS_INT32_TYPE_
# error OBS_INT32_TYPE_ is not defined.
#endif

#ifndef OBS_UINT32_TYPE_
# error OBS_UINT32_TYPE_ is not defined.
#endif

#ifndef OBS_PAGE_OFFSET_TYPE_
/**@internal smallest unsigned type enough for page offset. */
# define OBS_PAGE_OFFSET_TYPE_		OBS_u16
# ifndef OBS_PAGE_OFFSET_ALIGN_
/**@internal shift amount of byte alignment for page offset type. */
#  define OBS_PAGE_OFFSET_ALIGN_	1
# endif
# ifndef OBS_PAGE_OFFSET_MAX_
/**@internal max offset value. */
#  define OBS_PAGE_OFFSET_MAX_		0xFFFFU
# endif
#endif

#endif
