/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**@file */

#ifndef OBS_DEFINES_H_
#define OBS_DEFINES_H_

#include <obs/obs_config.h>

/**@internal begin of declaration of functions. */
#define OBS_BEGIN_DECL_		OBS_BEGIN_DECL__
/**@internal end of declaration of functions. */
#define OBS_END_DECL_		OBS_END_DECL__

/**@internal define an inline function. */
#define OBS_DEF_INLINE_FUNC_(ret, name, args)				      \
	OBS_DEF_INLINE_FUNC__(ret, name, args)

/**@internal define an inline function having unused argument(s). */
#define OBS_DEF_INLINE_FUNC_AU_(ret, name, args)			      \
	OBS_DEF_INLINE_FUNC_AU__(ret, name, args)

/**@internal initializer for the structure element. */
#define OBS_STRUCT_INIT_(name, value)					      \
	OBS_STRUCT_INIT__(name, value)

/**@internal initializer for an array element. */
#define OBS_ARRAY_INIT_(suf, value)					      \
	OBS_ARRAY_INIT__(suf, value)

/**@internal get number of array elements. */
#define OBS_NUM_OF(array)						      \
	(sizeof (array) / sizeof (array[0]))

/**
 * shift amount of least value fo max alignment which is ensured
 * under all platforms.
 */
#define OBS_LEAST_MAX_ALIGNMENT		8

/** shift amont of 1 byte alignment. */
#define OBS_ALIGN_1		0
/** shift amont of 2 bytes (word) alignment. */
#define OBS_ALIGN_2		1
/** shift amont of 4 bytes (double word) alignment. */
#define OBS_ALIGN_4		2
/** shift amont of 8 bytes (quad word) alignment. */
#define OBS_ALIGN_8		3
/** shift amont of 16 bytes alignment. */
#define OBS_ALIGN_16		4
/** shift amont of 32 bytes alignment. */
#define OBS_ALIGN_32		5
/** shift amont of 64 bytes alignment. */
#define OBS_ALIGN_64		6
/** shift amont of 128 bytes alignment. */
#define OBS_ALIGN_128		7
/** shift amont of 256 bytes alignment. */
#define OBS_ALIGN_256		8
/** shift amount of specified type. */
#define OBS_ALIGN_TYPE(type)						      \
	((/*CONSTCOND*/sizeof (type) >= 2)+				      \
	 (/*CONSTCOND*/sizeof (type) >= 4)+				      \
	 (/*CONSTCOND*/sizeof (type) >= 8)+				      \
	 (/*CONSTCOND*/sizeof (type) >= 16)+				      \
	 (/*CONSTCOND*/sizeof (type) >= 32)+				      \
	 (/*CONSTCOND*/sizeof (type) >= 64)+				      \
	 (/*CONSTCOND*/sizeof (type) >= 128)+				      \
	 (/*CONSTCOND*/sizeof (type) >= 256))
/** shift amount of pointer. */
#ifdef OBS_ALIGN_POINTER_
#define OBS_ALIGN_POINTER	OBS_ALIGN_POINTER_
#else
#define OBS_ALIGN_POINTER	OBS_ALIGN_TYPE(void *)
#endif

#endif
