/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef OBS_PA_INSP_FILE_H_
#define OBS_PA_INSP_FILE_H_

/**@file */

#include <obs/obs_pa_inspector.h>
#include <stdio.h>

/** inspector file class object. */
typedef struct OBS_pa_inspector_file OBS_pa_inspector_file;

/** inspector file class object. */
struct OBS_pa_inspector_file
{
	/** file object to which inspection result is stored. */
	FILE *fp;
	/** whether the last output is ended by newline. */
	OBS_bool newline;
	/** each line prefix of the output. */
	const char *prefix;
	/** inspector class object. */
	OBS_pa_inspector inspector;
};

OBS_BEGIN_DECL_
/**
 * initialize inspector file class object.
 *
 * @param newops	class object to be initialzed.
 * @param insp_ops	class object to be inspected.
 * @param fp		file object to which inspection result is stored.
 * @param prefix	each line prefix of the output.
 * @return		base class object of @a newops.
 */
const OBS_pa_ops *obs_inspect_pa_ops_file(
	OBS_pa_inspector_file *newops,
	const OBS_pa_ops *insp_ops,
	FILE *fp, const char *prefix);
OBS_END_DECL_

#endif
