/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**@file */

#ifndef OBS_PA_INSPECTOR_H_
#define OBS_PA_INSPECTOR_H_

#include <obs/obs_pa_wrapper.h>

/** inspector class object. */
typedef struct OBS_pa_inspector OBS_pa_inspector;

/** pointer to an output function. */
typedef void (*OBS_pa_inspector_output_func)(void *, const char *, ...);

/** inspector class object. */
struct OBS_pa_inspector
{
	/**@internal output function. */
	OBS_pa_inspector_output_func output_func;
	/**@internal user data for output_func. */
	void *closure;
	/**@internal wrapper class object. */
	OBS_pa_wrapper wrapper;
};

OBS_BEGIN_DECL_
/**
 * output the wrapper event in the human-readable form.
 *
 * @param output_func	output function.
 * @param closure	user data for output_func.
 * @param event		event to be shown.
 */
void obs_pa_inspector_show_wrapper_event(
	OBS_pa_inspector_output_func output_func,
	void *closure, const OBS_pa_wrapper_event *event);
/**
 * initialize inspector class object.
 *
 * @param newops	inspector class object to be initialized.
 * @param insp_ops	allocator class object to be inspected.
 * @param output_func	output function.
 * @param closure	user data for output_func.
 * @return		base class object of @a newops.
 */
const OBS_pa_ops *obs_inspect_pa_ops(
	OBS_pa_inspector *newops, const OBS_pa_ops *insp_ops,
	OBS_pa_inspector_output_func output_func, void *closure);
OBS_END_DECL_

#endif
