/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef OBS_SLAB_PRIV_
#error OBS_SLAB_PRIV_ is not defined.
#endif
#ifndef OBS_SLAB_DATA_
#error OBS_SLAB_DATA_ is not defined.
#endif
#ifndef OBS_SLAB_QUEUE_
#error OBS_SLAB_QUEUE_ is not defined.
#endif
#ifndef OBS_SLAB_OFFSET_
#error OBS_SLAB_OFFSET_ is not defined.
#endif
#ifndef OBS_SLAB_OFFSET_MAX_
#error OBS_SLAB_OFFSET_MAX_ is not defined.
#endif
#ifndef OBS_SLAB_OFFSET_ALIGN_
#error OBS_SLAB_OFFSET_ALIGN_ is not defined.
#endif
#ifndef OBS_SLAB_LINK_TERM_
#error OBS_SLAB_LINK_TERM_ is not defined.
#endif
#ifndef OBS_SLAB_LINK_INUSE_
#error OBS_SLAB_LINK_INUSE_ is not defined.
#endif
#ifndef OBS_SLAB_OBJ_TO_SLAB_HEAD_
#error OBS_SLAB_OBJ_TO_SLAB_HEAD_ is not defined.
#endif
#ifndef OBS_SLAB_INIT_ALLOCATOR_
#error OBS_SLAB_INIT_ALLOCATOR_ is not defined.
#endif
#ifndef OBS_SLAB_INIT_BACKEND_ALLOCATOR_
#error OBS_SLAB_INIT_BACKEND_ALLOCATOR_ is not defined.
#endif
#ifndef OBS_SLAB_UNINIT_BACKEND_ALLOCATOR_
#error OBS_SLAB_UNINIT_BACKEND_ALLOCATOR_ is not defined.
#endif
#ifndef OBS_SLAB_ALLOCATE_RAW_SLAB_
#error OBS_SLAB_ALLOCATE_RAW_SLAB_ is not defined.
#endif
#ifndef OBS_SLAB_FREE_RAW_SLAB_
#error OBS_SLAB_FREE_RAW_SLAB_ is not defined.
#endif

#ifdef OBS_SLAB_IMPL_PRE_

#include <obs/obs_slab.h>

typedef struct OBS_SLAB_PRIV_ OBS_SLAB_PRIV_;
typedef struct OBS_SLAB_DATA_ OBS_SLAB_DATA_;
typedef struct OBS_SLAB_QUEUE_ OBS_SLAB_QUEUE_;

#define OBS_SLAB_CALL_HOOK_NOCHECK_(slab, method, obj)			      \
	(*(slab)->hooks_->method)((slab), (slab)->hooks_priv_, (obj))
#define OBS_SLAB_CALL_HOOK_(slab, method, obj)				      \
	(((slab)->hooks_ && (slab)->hooks_->method) ?			      \
	 OBS_SLAB_CALL_HOOK_NOCHECK_((slab), method, (obj)) :		      \
	 OBS_S_NOERR)

OBS_DEF_INLINE_FUNC_(void, obs_slab_run_gc_if_need_, (OBS_slab *slab))
{
	if (slab->hiwat_ && slab->num_free_objects_ >= slab->hiwat_)
		obs_slab_run_gc(slab, slab->lowat_);
}


struct OBS_SLAB_QUEUE_
{
	OBS_SLAB_DATA_	*head;
};

struct OBS_SLAB_PRIV_
{
	OBS_SLAB_QUEUE_		active;
	OBS_SLAB_QUEUE_		empty;
	OBS_SLAB_QUEUE_		full;
	OBS_u8			page_shift;
};

/*
 * address map of a slab:
 *
 *                (lower address)
 *  slab head -> +---------------+  ---        ---
 *               |  pre_gap      |   | top      |
 *  object0 ---> ++-------------++  ---         |
 *               ||             ||   |          |
 *               ||entity of obj||   |          |
 *               ||             ||   |          |
 *               |+-------------+|   |          |
 *               || pad         ||   |          |
 *               |+-------------+|   | elemsize |
 *               || link        ||   |          |
 *               |+-------------+|   |          |
 *               ||             ||   |          |
 *               || pad         ||   |          |
 *               ||             ||   |          |
 *  object1 ---> ++-------------++  ---         |
 *               ||             ||   |          |
 *               ||entity of obj||   |          |
 *               ||             ||   |          |
 *               |+-------------+|   |          | slab size
 *               || pad         ||   |          |
 *               |+-------------+|   | elemsize |
 *               || link        ||   |          |
 *               |+-------------+|   |          |
 *               ||             ||   |          |
 *               || pad         ||   |          |
 *               ||             ||   |          |
 *               ++-------------++  ---         |
 *                 .....
 *               +---------------+              |
 *               |  post_gap     |              |
 *               +---------------+              |
 *               |  slab data    |              |
 *               |+-------------+|              |
 *               || next_slab   ||              |
 *               || pprev_slab  ||              |
 *               || uninit_list ||              |
 *               || unused_list ||              |
 *               || num_in_use  ||              |
 *               |+-------------+|              |
 *               +---------------+             ---
 *
 *  - pre_gap size is variable in respective slabs, selected from
 *    the values satisfying the alignment condition of objects.
 *  - each object has one of three states: uninit, unused or inuse.
 *    uninit or unused objects are linked by uninit_list or
 *    unused_list, respectively, and link field.
 *  - link fields is placed at the tail of each objects, and
 *    this points to the next object which has the same state.
 *    To get the pointer of the next object, apply the following expression:
 *      next object = slab head + link
 *  - Inuse object is not linked, because of preventing unlinking.
 *    (note: cost to unlink any element from single linked list is O(n).)
 *    Instead, LINK_INUSE is stored to link.
 *
 */

struct OBS_SLAB_DATA_
{
	OBS_SLAB_DATA_		*next_slab;
	OBS_SLAB_DATA_		**pprev_slab;
	OBS_SLAB_OFFSET_	uninit_list;
	OBS_SLAB_OFFSET_	unused_list;
	OBS_SLAB_OFFSET_	top;
	OBS_SLAB_OFFSET_	num_in_use;
};

/* hash pointer */
OBS_DEF_INLINE_FUNC_(OBS_u32, hash, (OBS_uintptr ptr))
{
	int i;
	OBS_u32 hashval = 0, tmp;

	/* XXX: fixme. */
	for (i=0; i<sizeof (ptr); i++) {
		hashval <<= 4;
		hashval += ptr & 0xFF;
		tmp = hashval & 0xF0000000;
		hashval ^= tmp;
		hashval ^= tmp >> 24;
		ptr >>= 8;
	}
	return hashval;
}


/* ----------------------------------------------------------------------
 * slab translation.
 */

OBS_DEF_INLINE_FUNC_(OBS_size, roundup, (OBS_size size, int align))
{
	OBS_size mask = (1 << align)-1;
	return (size + mask) & ~mask;
}

OBS_DEF_INLINE_FUNC_(int, get_page_shift, (OBS_slab *slab))
{
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;
	return priv->page_shift;
}

OBS_DEF_INLINE_FUNC_(OBS_size, get_page_size, (OBS_slab *slab))
{ return (OBS_size)1 << get_page_shift(slab); }

OBS_DEF_INLINE_FUNC_(void *, add_ptr, (void *ptr, OBS_size ofs))
{ return (char *)ptr + ofs; }

OBS_DEF_INLINE_FUNC_(void *, sub_ptr, (void *ptr, OBS_size ofs))
{ return (char *)ptr - ofs; }

OBS_DEF_INLINE_FUNC_(OBS_SLAB_DATA_ *, slab_head_to_data,
		     (OBS_slab *slab, void *head))
{ return add_ptr(head, slab->slab_size_ - sizeof (OBS_SLAB_DATA_)); }

OBS_DEF_INLINE_FUNC_(void *, data_to_slab_head,
		     (OBS_slab *slab, OBS_SLAB_DATA_ *data))
{ return sub_ptr(data + 1, slab->slab_size_); }

OBS_DEF_INLINE_FUNC_(void *, ofs_to_ptr,
		     (OBS_slab *slab, OBS_SLAB_DATA_ *data,
		      OBS_SLAB_OFFSET_ ofs))
{
	return add_ptr(data_to_slab_head(slab, data), ofs);
}

OBS_DEF_INLINE_FUNC_(OBS_SLAB_OFFSET_ *, obj_to_link,
		     (OBS_slab *slab, void *obj))
{ return add_ptr(obj, slab->object_size_); }

OBS_DEF_INLINE_FUNC_(OBS_SLAB_OFFSET_, to_ofs, (void *head, void *ptr))
{ return (OBS_SLAB_OFFSET_)((char *)ptr - (char *)head); }


/* ----------------------------------------------------------------------
 * queue manipulation.
 */

OBS_DEF_INLINE_FUNC_(void, queue_init, (OBS_SLAB_QUEUE_ *queue))
{ queue->head = NULL; }

OBS_DEF_INLINE_FUNC_(void, queue_insert_head,
		     (OBS_SLAB_QUEUE_ *queue, OBS_SLAB_DATA_ *data))
{
	if (queue->head)
		queue->head->pprev_slab = &data->next_slab;
	data->next_slab = queue->head;
	data->pprev_slab = &queue->head;
	queue->head = data;
}

OBS_DEF_INLINE_FUNC_(void, queue_remove, (OBS_SLAB_DATA_ *data))
{
	if (data->next_slab)
		data->next_slab->pprev_slab = data->pprev_slab;
	*data->pprev_slab = data->next_slab;
}

OBS_DEF_INLINE_FUNC_(OBS_bool, queue_is_empty, (OBS_SLAB_QUEUE_ *queue))
{ return queue->head == NULL; }

OBS_DEF_INLINE_FUNC_(OBS_SLAB_DATA_ *, queue_head, (OBS_SLAB_QUEUE_ *queue))
{ return queue->head; }

OBS_DEF_INLINE_FUNC_(OBS_SLAB_DATA_ *, queue_next, (OBS_SLAB_DATA_ *data))
{ return data->next_slab; }

OBS_DEF_INLINE_FUNC_(void, queue_insert_next,
		     (OBS_SLAB_DATA_ *base, OBS_SLAB_DATA_ *newslab))
{
	/* assumption: base != NULL */
	if (base->next_slab)
		base->next_slab->pprev_slab = &newslab->next_slab;
	newslab->next_slab = base->next_slab;
	base->next_slab = newslab;
	newslab->pprev_slab = &base->next_slab;
}

#undef OBS_SLAB_IMPL_PRE_
#endif /* OBS_SLAB_IMPL_PRE_ */

#ifdef OBS_SLAB_IMPL_POST_

/* ----------------------------------------------------------------------
 * allocator template
 */

static OBS_status
new_slab(OBS_slab *slab, OBS_SLAB_DATA_ **rdata)
{
	OBS_status ret;
	OBS_size top, num_objects;
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;
	OBS_SLAB_DATA_ *data;
	OBS_SLAB_OFFSET_ *plink;
	void *head;

	ret = OBS_SLAB_ALLOCATE_RAW_SLAB_(slab, &head);
	if (ret)
		return ret;
	data = *rdata = slab_head_to_data(slab, head);

	/*
	 * determine pre_gap.
	 *
	 * assumption:
	 *   - slab->alignment_ is well-aligned both for given alignment
	 *     value and for OBS_SLAB_OFFSET_.
	 *   - object size is well-aligned for OBS_SLAB_OFFSET_.
	 *   - slab->num_ranks_of_top > 0
	 */
	top = (hash((OBS_uintptr)head >> get_page_shift(slab)) %
	       slab->num_ranks_of_top_) << slab->alignment_;
	data->top = data->uninit_list = top;
	data->unused_list = OBS_SLAB_LINK_TERM_;
	data->num_in_use = 0;
	slab->num_free_objects_ += slab->num_objects_per_slab_;

	/* fill free list. */
	plink = obj_to_link(slab, ofs_to_ptr(slab, data, top));
	num_objects = slab->num_objects_per_slab_;
	while (--num_objects>0) {
		top += slab->element_size_;
		*plink = top;
		plink = add_ptr(plink, slab->element_size_);
	}
	*plink = OBS_SLAB_LINK_TERM_;

	return OBS_S_NOERR;
}

static void
free_slab(OBS_slab *slab, OBS_SLAB_DATA_ *data)
{
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;
	void *ptr;
	OBS_SLAB_OFFSET_ cur, i;

	/* step1: destruct all unused entries */
	if (slab->hooks_ && slab->hooks_->destruct) {
		cur = data->unused_list;
		while (cur != OBS_SLAB_LINK_TERM_) {
			ptr = ofs_to_ptr(slab, data, cur);
			cur = *obj_to_link(slab, ptr);
			OBS_SLAB_CALL_HOOK_NOCHECK_(slab, destruct, ptr);
		}
	}
	/* step2: put and destruct all inuse objects. */
	if (slab->hooks_ && (slab->hooks_->put || slab->hooks_->destruct) &&
	    data->num_in_use) {
		ptr = ofs_to_ptr(slab, data, data->top);
		for (i=slab->num_objects_per_slab_; i>0; i--) {
			if (*obj_to_link(slab, ptr) == OBS_SLAB_LINK_INUSE_) {
				OBS_SLAB_CALL_HOOK_NOCHECK_(slab, put, ptr);
				OBS_SLAB_CALL_HOOK_NOCHECK_(
					slab, destruct, ptr);
			}
			ptr = add_ptr(ptr, slab->element_size_);
		}
	}

	OBS_SLAB_FREE_RAW_SLAB_(slab, data_to_slab_head(slab, data));
}

static void
put_all_objects_in_slab(OBS_slab *slab, OBS_SLAB_DATA_ *data)
{
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;
	void *ptr, *pagehead;
	OBS_SLAB_OFFSET_ i, cur, prev, *plink;

	if (data->num_in_use) {
		cur = data->top;
		ptr = ofs_to_ptr(slab, data, data->top);
		prev = data->unused_list;
		for (i=slab->num_objects_per_slab_; i>0; i--) {
			plink = obj_to_link(slab, ptr);
			if (*plink == OBS_SLAB_LINK_INUSE_) {
				OBS_SLAB_CALL_HOOK_(slab, put, ptr);
				*plink = prev;
				prev = cur;
			}
			cur += slab->element_size_;
			ptr = add_ptr(ptr, slab->element_size_);
		}
		data->unused_list = prev;
		slab->num_free_objects_ += data->num_in_use;
		data->num_in_use = 0;
	}
}


static const OBS_slab_ops_ slabOps;

OBS_status
OBS_SLAB_INIT_ALLOCATOR_(OBS_slab *slab, OBS_size objsize, OBS_flags features,
			 int align,
			 const OBS_slab_hooks *hooks, void *hooks_priv,
			 const OBS_pa_ops *paops)
{
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;
	OBS_status ret;
	OBS_size poolsize, wastage, elemsize, num_objects;

	priv->page_shift = obs_pa_ops_get_alignment(paops);
	slab->num_slabs_ = 0;
	slab->num_free_objects_ = 0;

	/*
	 * calculate and mormalize some amounts.
	 *
	 * elem =  object + pad1 + link + pad2
	 *
	 * pad1 : align link field with its natural alignment.
	 * pad2 : align next object with the value specified by caller.
	 *
	 */
	/* To unify pad1 length, carry align value. */
	align = align>OBS_SLAB_OFFSET_ALIGN_ ? align : OBS_SLAB_OFFSET_ALIGN_;
	slab->alignment_ = align;
	/* To simplify, add pad1 size to objsize */
	objsize = roundup(objsize, OBS_SLAB_OFFSET_ALIGN_);
	slab->object_size_ = objsize;
	/* element size (see below) */
	elemsize = roundup(objsize + sizeof (OBS_SLAB_OFFSET_), align);
	slab->element_size_ = (OBS_SLAB_OFFSET_)elemsize;

	ret = OBS_SLAB_INIT_BACKEND_ALLOCATOR_(slab, features, paops);
	if (ret)
		return ret;

	poolsize = slab->slab_size_ - sizeof (OBS_SLAB_DATA_);
	/* wastage size per slab. (= pre_gap + post_gap) */
	wastage = poolsize % elemsize;
	num_objects = poolsize / elemsize;

	slab->hooks_ = hooks;
	slab->hooks_priv_ = hooks_priv;
	slab->ops_ = &slabOps;
	slab->num_objects_per_slab_ = (OBS_SLAB_OFFSET_)num_objects;
	slab->wastage_per_slab_ = wastage;
	slab->num_ranks_of_top_ = (OBS_SLAB_OFFSET_)(wastage >> align) + 1;
	queue_init(&priv->active);
	queue_init(&priv->empty);
	queue_init(&priv->full);

	return OBS_S_NOERR;
}

static OBS_status
get_object(OBS_slab *slab, void **robj)
{
	OBS_status ret;
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;
	OBS_SLAB_DATA_ *data;
	OBS_SLAB_OFFSET_ objofs, *plink;
	void *obj;

	/* pick a slab having an unused or an uninit entry. */
	if (!queue_is_empty(&priv->active)) {
		data = queue_head(&priv->active);
		queue_remove(data);
	} else if (!queue_is_empty(&priv->empty)) {
		data = queue_head(&priv->empty);
		queue_remove(data);
	} else {
		ret = new_slab(slab, &data);
		if (ret)
			return ret;
	}

	/* get an object. */
	if (data->unused_list != OBS_SLAB_LINK_TERM_) {
		objofs = data->unused_list;
		obj = ofs_to_ptr(slab, data, objofs);
		plink = obj_to_link(slab, obj);
		data->unused_list = *plink;
	} else {
		objofs = data->uninit_list;
		obj = ofs_to_ptr(slab, data, objofs);
		plink = obj_to_link(slab, obj);
		data->uninit_list = *plink;
		ret = OBS_SLAB_CALL_HOOK_(slab, construct, obj);
		if (ret) {
			*plink = data->uninit_list;
			data->uninit_list = objofs;
			goto err_by_hooks;
		}
	}
	ret = OBS_SLAB_CALL_HOOK_(slab, get, obj);
	if (ret) {
		*plink = data->unused_list;
		data->unused_list = objofs;
		goto err_by_hooks;
	}
	data->num_in_use ++;
	slab->num_free_objects_ --;
	*plink = OBS_SLAB_LINK_INUSE_;

	*robj = obj;
	ret = OBS_S_NOERR;

err_by_hooks:
	/* place the slab in an appropriate queue. */
	if (data->num_in_use == 0)
		queue_insert_head(&priv->empty, data);
	else if (data->num_in_use == slab->num_objects_per_slab_)
		queue_insert_head(&priv->full, data);
	else
		queue_insert_head(&priv->active, data);

	return ret;
}

static void
put_object(OBS_slab *slab, void *obj)
{
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;
	void *slab_head;
	OBS_SLAB_DATA_ *data, *curdata, *nextdata;
	OBS_SLAB_OFFSET_ *plink;

	slab_head = OBS_SLAB_OBJ_TO_SLAB_HEAD_(slab, obj);
	if (slab_head == NULL) {
		/* XXX: unknown object. */
		return;
	}

	data = slab_head_to_data(slab, slab_head);
	plink = obj_to_link(slab, obj);

	if (*plink != OBS_SLAB_LINK_INUSE_) {
		/* XXX: possibly double free or buffer overrun. */
		return;
	}
	*plink = data->unused_list;
	data->unused_list = to_ofs(slab_head, obj);
	OBS_SLAB_CALL_HOOK_(slab, put, obj);
	slab->num_free_objects_ ++;
	data->num_in_use --;

	if (data->num_in_use == 0) {
		queue_remove(data);
		queue_insert_head(&priv->empty, data);
	} else {
		/* adjust priority. */
		curdata = data;
		while ((nextdata = queue_next(curdata)) != NULL) {
			if (nextdata->num_in_use <= data->num_in_use)
				break;
			curdata = nextdata;
		}
		if (curdata != data) {
			queue_remove(data);
			queue_insert_next(curdata, data);
		}
	}

	obs_slab_run_gc_if_need_(slab);
}

static void
put_all_objects_in_link(OBS_slab *slab, OBS_SLAB_QUEUE_ *queue)
{
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;
	OBS_SLAB_DATA_ *data;
	while (!queue_is_empty(queue)) {
		data = queue_head(queue);
		queue_remove(data);
		put_all_objects_in_slab(slab, data);
		queue_insert_head(&priv->empty, data);
	}
}

static void
put_all_objects(OBS_slab *slab)
{
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;
	put_all_objects_in_link(slab, &priv->full);
	put_all_objects_in_link(slab, &priv->active);
	obs_slab_run_gc_if_need_(slab);
}

static void
free_all_slabs_in_link(OBS_slab *slab, OBS_SLAB_QUEUE_ *queue)
{
	OBS_SLAB_DATA_ *data;
	while (!queue_is_empty(queue)) {
		data = queue_head(queue);
		queue_remove(data);
		free_slab(slab, data);
	}
}

static void
uninit_allocator(OBS_slab *slab)
{
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;

	free_all_slabs_in_link(slab, &priv->full);
	free_all_slabs_in_link(slab, &priv->active);
	free_all_slabs_in_link(slab, &priv->empty);
	OBS_SLAB_UNINIT_BACKEND_ALLOCATOR_(slab);
}

static void
run_gc(OBS_slab *slab, OBS_size lowat)
{
	OBS_SLAB_PRIV_ *priv = (void *)&slab->ops_priv_;
	OBS_SLAB_DATA_ *data;

	while (!queue_is_empty(&priv->empty) &&
	       slab->num_free_objects_>lowat) {
		data = queue_head(&priv->empty);
		queue_remove(data);
		free_slab(slab, data);
		slab->num_free_objects_ -= slab->num_objects_per_slab_;
	}
}

static const OBS_slab_ops_ slabOps = {
	OBS_STRUCT_INIT_(uninit, uninit_allocator),
	OBS_STRUCT_INIT_(get, get_object),
	OBS_STRUCT_INIT_(put, put_object),
	OBS_STRUCT_INIT_(put_all, put_all_objects),
	OBS_STRUCT_INIT_(run_gc, run_gc)
};

#undef OBS_SLAB_IMPL_POST_
#endif /* OBS_SLAB_IMPL_POST_ */
