/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define OBS_SLAB_PRIV_				OBS_slab_large_priv_base_
#define OBS_SLAB_DATA_				OBS_slab_large_data_
#define OBS_SLAB_QUEUE_				OBS_slab_large_queue_
#define OBS_SLAB_OFFSET_			OBS_u32
#define OBS_SLAB_OFFSET_MAX_			(0xFFFFFFFFU)
#define OBS_SLAB_OFFSET_ALIGN_			2
#define OBS_SLAB_LINK_TERM_			OBS_PAGE_OFFSET_MAX
#define OBS_SLAB_LINK_INUSE_			(OBS_PAGE_OFFSET_MAX-1)
#define OBS_SLAB_OBJ_TO_SLAB_HEAD_		obj_to_slab_head
#define OBS_SLAB_INIT_ALLOCATOR_		obs_slab_large_init_
#define OBS_SLAB_INIT_BACKEND_ALLOCATOR_	init_backend_allocator
#define OBS_SLAB_UNINIT_BACKEND_ALLOCATOR_	uninit_backend_allocator
#define OBS_SLAB_ALLOCATE_RAW_SLAB_		allocate_raw_slab
#define OBS_SLAB_FREE_RAW_SLAB_			free_raw_slab

#define OBS_SLAB_IMPL_PRE_
#include <obs/obs_slab_impl.h>

typedef struct OBS_slab_large_priv_ OBS_slab_large_priv_;
typedef struct OBS_slab_large_extra_ OBS_slab_large_extra_;
typedef struct OBS_slab_large_hash_prime_ OBS_slab_large_hash_prime_;

struct OBS_slab_large_priv_
{
	OBS_slab_large_priv_base_	base;
	void				*hash_table;
	OBS_u32				num_pages_per_slab;
	OBS_u32				hash_rank;
};

struct OBS_slab_large_hash_prime_
{
	OBS_u32		table_size;
	OBS_u32		soft_limit;
	OBS_u32		modulo;
	OBS_u32		rehash;
};

static const OBS_slab_large_hash_prime_ hash_primes[] = {
	{         32U,         28U,         31U,         29U },
	{         64U,         55U,         61U,         59U },
	{        128U,         98U,        109U,        107U },
	{        256U,        217U,        241U,        239U },
	{        512U,        417U,        463U,        461U },
	{       1024U,        919U,       1021U,       1019U },
	{       2048U,       1826U,       2029U,       2027U },
	{       4096U,       3684U,       4093U,       4091U },
	{       8192U,       7280U,       8089U,       8087U },
	{      16384U,      14727U,      16363U,      16361U },
	{      32768U,      29447U,      32719U,      32717U },
	{      65536U,      58969U,      65521U,      65519U },
	{     131072U,     117910U,     131011U,     131009U },
	{     262144U,     235900U,     262111U,     262109U },
	{     524288U,     471799U,     524221U,     524219U },
	{    1048576U,     943716U,    1048573U,    1048571U },
	{    2097152U,    1887420U,    2097133U,    2097131U },
	{    4194304U,    3774423U,    4193803U,    4193801U },
	{    8388608U,    7549606U,    8388451U,    8388449U },
	{   16777216U,   15099427U,   16777141U,   16777139U },
	{   33554432U,   30198610U,   33554011U,   33554009U },
	{   67108864U,   60397802U,   67108669U,   67108667U },
	{  134217728U,  120795695U,  134217439U,  134217437U },
	{  268435456U,  241591508U,  268435009U,  268435007U },
	{  536870912U,  483183755U,  536870839U,  536870837U },
	{ 1073741824U,  966367547U, 1073741719U, 1073741717U },
	{ 2147483648U, 1932734656U, 2147482951U, 2147482949U },
};

static void
static_assersion(void)
{
	/* check whether OBS_slab_small_priv_ can go in ops_priv_ field. */
	char remain[sizeof (((OBS_slab *)0)->ops_priv_) >=
		    sizeof (OBS_slab_large_priv_) ? 1:-1];
	(void)remain;
}

#define HASH_NOT_FOUND	(~(OBS_u32)0)
#define HASH_FREE_PAGE	(~(OBS_u32)0)

#define HASH_FIND_(name, cond)						      \
static OBS_u32								      \
name(OBS_slab *slab, OBS_uintptr page)					      \
{									      \
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;		      \
	OBS_uintptr *hash_table = priv->hash_table;			      \
	OBS_u32 hashval, index;						      \
	OBS_uintptr slabhead;						      \
									      \
	hashval = index =						      \
	    hash(page >> get_page_shift(slab)) %			      \
	    hash_primes[priv->hash_rank].modulo;			      \
									      \
	do {								      \
		slabhead = hash_table[index];				      \
		if ((cond))						      \
			return index;					      \
		index += hash_primes[priv->hash_rank].rehash;		      \
		if (index >= hash_primes[priv->hash_rank].modulo)	      \
			index -= hash_primes[priv->hash_rank].modulo;	      \
	} while (index != hashval);					      \
									      \
	return HASH_NOT_FOUND;						      \
}

HASH_FIND_(find_page_from_hash,
	   slabhead >= page && slabhead <= page+slab->slab_size_-1)
HASH_FIND_(find_free_space_from_hash, slabhead == 0)

static void *
obj_to_slab_head(OBS_slab *slab, void *obj)
{
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	OBS_uintptr *hash_table = priv->hash_table;
	OBS_uintptr page, slabhead;
	OBS_u32 index;

	page = (OBS_uintptr)obj & ~(get_page_size(slab)-1);
	if (hash_table == NULL)
		return (void *)page;

	index = find_page_from_hash(slab, page);
	if (index == HASH_NOT_FOUND)
		return NULL;
	return (void *)hash_table[index];
}

static void
add_page_to_hash_noresize(OBS_slab *slab,
			  OBS_uintptr slabhead, OBS_uintptr page)
{
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	OBS_uintptr *hash_table = priv->hash_table;
	OBS_u32 index;

	index = find_free_space_from_hash(slab, page);

	if (index == HASH_NOT_FOUND)
		/* XXX: assert */
		return;

	hash_table[index] = slabhead;
}

static void
remove_page_from_hash_noresize(OBS_slab *slab,
			       OBS_uintptr slabhead, OBS_uintptr page)
{
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	OBS_uintptr *hash_table = priv->hash_table;
	OBS_u32 index;

	index = find_page_from_hash(slab, page);

	if (index == HASH_NOT_FOUND)
		/* XXX: assert */
		return;

	hash_table[index] = HASH_FREE_PAGE;
}

static void
do_slab_pages(OBS_slab *slab, OBS_uintptr slabhead,
	      void (*func)(OBS_slab *slab, OBS_uintptr, OBS_uintptr))
{
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	OBS_u32 i;
	OBS_uintptr pagehead;
	OBS_size pagesize = get_page_size(slab);

	pagehead = slabhead;
	for (i=priv->num_pages_per_slab; i>0; i--) {
		(*func)(slab, slabhead, pagehead);
		pagehead += pagesize;
	}
}

static void
add_slab_to_hash_noresize(OBS_slab *slab, OBS_uintptr slabhead)
{
	do_slab_pages(slab, slabhead, &add_page_to_hash_noresize);
}

static void
add_slab_queue_to_hash_noresize(OBS_slab *slab, OBS_slab_large_queue_ *queue)
{
	OBS_slab_large_data_ *data;

	for (data=queue_head(queue); data; data=queue_next(data))
		add_slab_to_hash_noresize(
			slab, (OBS_uintptr)data_to_slab_head(slab, data));
}

static void
remove_slab_from_hash_noresize(OBS_slab *slab, OBS_uintptr slabhead)
{
	do_slab_pages(slab, slabhead, &remove_page_from_hash_noresize);
}

OBS_DEF_INLINE_FUNC_(OBS_size, calc_hash_table_bytes,
		     (OBS_slab *slab, int rank))
{
	return hash_primes[rank].table_size * sizeof (OBS_uintptr);
}

OBS_DEF_INLINE_FUNC_(OBS_size, calc_hash_table_pages,
		     (OBS_slab *slab, int rank))
{
	OBS_size num_pages =
	    calc_hash_table_bytes(slab, rank) / get_page_size(slab);
	num_pages || num_pages++;
	return num_pages;
}

static OBS_status
alloc_hash_table(OBS_slab *slab, int rank, void **rhead)
{
	OBS_status ret;
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	OBS_region region;

	ret = obs_pa_alloc(&slab->pagealloc_, &region,
			   calc_hash_table_pages(slab, rank));
	if (ret)
		return ret;

	/* assumption: new pages are already zero-filled. */

	*rhead = region.head;

	return OBS_S_NOERR;
}

static void
free_hash_table(OBS_slab *slab)
{
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	OBS_region region;

	if (priv->hash_table) {
		region.size = calc_hash_table_bytes(slab, priv->hash_rank);
		region.head = priv->hash_table;
		obs_pa_free(&slab->pagealloc_, &region);
	}
}

static OBS_status
build_hash_table(OBS_slab *slab)
{
	OBS_status ret;
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	OBS_u32 index;

	/* assumption: hash table is already zero-filled. */

	add_slab_queue_to_hash_noresize(slab, &priv->base.active);
	add_slab_queue_to_hash_noresize(slab, &priv->base.full);
	/* XXX: empty queue should be empty, though. */
	add_slab_queue_to_hash_noresize(slab, &priv->base.empty);
}

static int
get_base_rank(OBS_slab *slab)
{
	OBS_size num_entries = get_page_size(slab) / sizeof (OBS_uintptr);
	int i;

	for (i=0; i<(int)OBS_NUM_OF(hash_primes); i++)
		if (hash_primes[i].table_size >= num_entries)
			return i;

	/* XXX: assert */
	return 0;
}

static OBS_status
init_hash_table(OBS_slab *slab)
{
	OBS_status ret;
	int rank;
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;

	rank = get_base_rank(slab);
	ret = alloc_hash_table(slab, rank, &priv->hash_table);
	if (ret)
		return ret;
	priv->hash_rank = rank;

	return OBS_S_NOERR;
}

static OBS_status
resize_hash_table(OBS_slab *slab, int newrank)
{
	OBS_status ret;
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	void *newtable;

	/* XXX: fixme to use expand/truncate */
	ret = alloc_hash_table(slab, newrank, &newtable);
	if (ret)
		return ret;

	free_hash_table(slab);
	priv->hash_table = newtable;
	priv->hash_rank = newrank;
	build_hash_table(slab);

	return OBS_S_NOERR;
}

static OBS_status
add_slab_to_hash(OBS_slab *slab, void *head)
{
	OBS_status ret;
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	int rank;

	rank = priv->hash_rank;
	if (slab->num_slabs_ >= hash_primes[rank].soft_limit) {
		rank++;
		if (rank >= OBS_NUM_OF(hash_primes))
			return OBS_S_NOMEM;
		ret = resize_hash_table(slab, rank);
		if (ret)
			return ret;
	}
	add_slab_to_hash_noresize(slab, (OBS_uintptr)head);

	return OBS_S_NOERR;
}

static void
remove_slab_from_hash(OBS_slab *slab, void *head)
{
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	int rank;

	remove_slab_from_hash_noresize(slab, (OBS_uintptr)head);

	rank = priv->hash_rank;
	/* to prevent congestion, delay resizing for one rank. */
	if (rank >= 2 && slab->num_slabs_ < hash_primes[rank-2].soft_limit &&
	    rank > get_base_rank(slab)) {
		rank--;
		resize_hash_table(slab, rank);
	}
}

/* hard limit to stop optimization of wastage. */
#define MAX_PACKING_SLAB_PAGES	32

static OBS_status
init_backend_allocator(OBS_slab *slab, OBS_flags features,
		       const OBS_pa_ops *paops)
{
	OBS_status ret;
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;
	const OBS_size pagesize = obs_pa_ops_get_page_unit(paops);
	OBS_size slabsize, elemsize;

	int align = slab->alignment_;
	elemsize = slab->element_size_;
	priv->hash_table = NULL;
	priv->num_pages_per_slab = 0;
	priv->hash_rank = 0;

	/* alignment cannot be over the page alignment. */
	if (align > obs_pa_ops_get_alignment(paops))
		return OBS_S_INVAL;

	slabsize = (elemsize + pagesize - 1) & ~(pagesize - 1);
	if (slabsize - sizeof (OBS_slab_large_data_) < elemsize)
		slabsize += pagesize;

	if (slabsize > MAX_PACKING_SLAB_PAGES*pagesize)
		features |= OBS_SLAB_F_FORCE_NO_HASH;

	if (!OBS_SLAB_FEATURES_IS_F(features, FORCE_NO_HASH)) {
		/* calc minimum wastage. */
		OBS_size wastage, min_wastage, min_slabsize, i;
		min_wastage = OBS_SIZE_MAX;
		for (i=slabsize>>get_page_shift(slab);
		     i<=MAX_PACKING_SLAB_PAGES; i++) {
			wastage = slabsize;
			wastage -= sizeof (OBS_slab_large_data_);
			wastage %= elemsize;
			if (wastage < min_wastage) {
				min_wastage = wastage;
				min_slabsize = slabsize;
			}
			slabsize += pagesize;
		}
		slabsize = min_slabsize;
		if (slabsize - sizeof (OBS_slab_large_data_) < (elemsize<<1))
			features |= OBS_SLAB_F_FORCE_NO_HASH;
	}
	slab->slab_size_ = slabsize;
	priv->num_pages_per_slab = slabsize >> get_page_shift(slab);

	ret = obs_pa_ops_init(paops, &slab->pagealloc_, 1, 0);
	if (ret)
		return ret;

	if (!OBS_SLAB_FEATURES_IS_F(features, FORCE_NO_HASH))
		return init_hash_table(slab);

	return OBS_S_NOERR;
}

OBS_DEF_INLINE_FUNC_(void, uninit_backend_allocator, (OBS_slab *slab))
{
	free_hash_table(slab);
	obs_pa_uninit(&slab->pagealloc_);
}

OBS_DEF_INLINE_FUNC_(OBS_status, allocate_raw_slab,
		     (OBS_slab *slab, void **head))
{
	OBS_status ret;
	OBS_region region;
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;

	ret = obs_pa_alloc(&slab->pagealloc_, &region,
			   priv->num_pages_per_slab);
	if (ret)
		return ret;

	slab->num_slabs_ ++;

	if (priv->hash_table) {
		ret = add_slab_to_hash(slab, region.head);
		if (ret) {
			obs_pa_free(&slab->pagealloc_, &region);
			return ret;
		}
	}

	*head = region.head;

	return OBS_S_NOERR;
}

OBS_DEF_INLINE_FUNC_(void, free_raw_slab, (OBS_slab *slab, void *head))
{
	OBS_region region;
	OBS_slab_large_priv_ *priv = (void *)&slab->ops_priv_;

	slab->num_slabs_ --;

	if (priv->hash_table)
		remove_slab_from_hash(slab, head);

	region.head = head;
	region.size = slab->slab_size_;

	obs_pa_free(&slab->pagealloc_, &region);
}

#define OBS_SLAB_IMPL_POST_
#include <obs/obs_slab_impl.h>
