/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define OBS_SLAB_PRIV_				OBS_slab_small_priv_
#define OBS_SLAB_DATA_				OBS_slab_small_data_
#define OBS_SLAB_QUEUE_				OBS_slab_small_queue_
#define OBS_SLAB_OFFSET_			OBS_page_offset
#define OBS_SLAB_OFFSET_MAX_			OBS_PAGE_OFFSET_MAX
#define OBS_SLAB_OFFSET_ALIGN_			OBS_PAGE_OFFSET_ALIGN
#define OBS_SLAB_LINK_TERM_			OBS_PAGE_OFFSET_MAX
#define OBS_SLAB_LINK_INUSE_			(OBS_PAGE_OFFSET_MAX-1)
#define OBS_SLAB_OBJ_TO_SLAB_HEAD_		obj_to_page_head
#define OBS_SLAB_INIT_ALLOCATOR_		obs_slab_small_init_
#define OBS_SLAB_INIT_BACKEND_ALLOCATOR_	init_backend_allocator
#define OBS_SLAB_UNINIT_BACKEND_ALLOCATOR_	uninit_backend_allocator
#define OBS_SLAB_ALLOCATE_RAW_SLAB_		allocate_raw_slab
#define OBS_SLAB_FREE_RAW_SLAB_			free_raw_slab

#define OBS_SLAB_IMPL_PRE_
#include <obs/obs_slab_impl.h>

static void
static_assersion(void)
{
	/* check whether OBS_slab_small_priv_ can go in ops_priv_ field. */
	char remain[sizeof (((OBS_slab *)0)->ops_priv_) >=
		    sizeof (OBS_slab_small_priv_) ? 1:-1];
	(void)remain;
}

OBS_DEF_INLINE_FUNC_(void *, obj_to_page_head, (OBS_slab *slab, void *obj))
{
	/*
	 * assumption:
	 *   block size is power of 2 and each block is aligned
	 *   to the block size.
	 */
	return (void *)((OBS_uintptr)obj & ~(get_page_size(slab)-1));
}

static OBS_status
init_backend_allocator(OBS_slab *slab, OBS_flags features,
		       const OBS_pa_ops *paops)
{
	OBS_size pagesize = get_page_size(slab);
	OBS_size poolsize = pagesize - sizeof (OBS_slab_small_data_);

	(void)&static_assersion;

	/* alignment cannot be over the page alignment. */
	if (slab->alignment_ > obs_pa_ops_get_alignment(paops))
		return OBS_S_INVAL;

	/* page size cannot be over the max value of page offset type. */
	if (pagesize > OBS_PAGE_OFFSET_MAX)
		return OBS_S_TOO_BIG;

	/* make sure that at least 1 objects can be housed in a slab. */
	if (slab->element_size_ > poolsize)
		return OBS_S_TOO_BIG;

	slab->slab_size_ = pagesize;

	return obs_pa_ops_init(paops, &slab->pagealloc_, 1, 0);
}

OBS_DEF_INLINE_FUNC_(void, uninit_backend_allocator, (OBS_slab *slab))
{
	obs_pa_uninit(&slab->pagealloc_);
}

OBS_DEF_INLINE_FUNC_(OBS_status, allocate_raw_slab,
		     (OBS_slab *slab, void **head))
{
	OBS_status ret;
	OBS_region page;

	ret = obs_pa_alloc(&slab->pagealloc_, &page, 1);
	if (ret)
		return ret;

	*head = page.head;

	slab->num_slabs_ ++;

	return OBS_S_NOERR;
}

OBS_DEF_INLINE_FUNC_(void, free_raw_slab, (OBS_slab *slab, void *head))
{
	OBS_region page;

	page.head = head;
	page.size = slab->slab_size_;

	obs_pa_free(&slab->pagealloc_, &page);

	slab->num_slabs_ --;
}

#define OBS_SLAB_IMPL_POST_
#include <obs/obs_slab_impl.h>
