/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/**@file */

#ifndef OBS_STATUS_H_
#define OBS_STATUS_H_

#include <obs/obs_types.h>
#include <obs/obs_defines.h>

/** result status code indicating no error. */
#define OBS_S_NOERR		0
/** result status code indicating memory was exhausted. */
#define OBS_S_NOMEM		1
/** result status code indicating it could not expand. */
#define OBS_S_CANNOT_EXPAND	2
/** result status code indicating invalid argument. */
#define OBS_S_INVAL		3
/** result status code indicating not supported. */
#define OBS_S_NOTSUPP		4
/** result status code indicating not contiguous region. */
#define OBS_S_NOT_CONTIGUOUS	5
/** result status code indicating too big. */
#define OBS_S_TOO_BIG		6
/** result status code indicating unexpected revocation. */
#define OBS_S_REVOKED		7
/** result status code indicating unknown error. */
#define OBS_S_UNKNOWN		8

OBS_BEGIN_DECL_
/**
 * convert result status code to the corresponding human-readable string.
 *
 * @param status	result status code.
 * @return		string corresponding to @a status.
 */
const char *obs_strstatus(OBS_status status);
OBS_END_DECL_

#endif
