/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <obs/obs.h>
#include <obs/obs_pa_insp_file.h>
#include <stdio.h>
#include <stdlib.h>

void
notice(const char *func, OBS_status errcode, int isexit)
{
	printf("%s failed: %s\n", func, obs_strstatus(errcode));
	if (isexit)
		exit(1);
}

OBS_status
construct(OBS_slab *slab, void *closure, void *obj)
{
	(void)slab;
	(void)closure;
	printf("  hooks> construct: %p\n", obj);
	return OBS_S_NOERR;
}

OBS_status
destruct(OBS_slab *slab, void *closure, void *obj)
{
	(void)slab;
	(void)closure;
	printf("  hooks> destruct: %p\n", obj);
	return OBS_S_NOERR;
}

OBS_status
get(OBS_slab *slab, void *closure, void *obj)
{
	(void)slab;
	(void)closure;
	printf("  hooks> get: %p\n", obj);
	return OBS_S_NOERR;
}

OBS_status
put(OBS_slab *slab, void *closure, void *obj)
{
	(void)slab;
	(void)closure;
	printf("  hooks> put: %p\n", obj);
	return OBS_S_NOERR;
}

OBS_slab_hooks hooks = {
	OBS_STRUCT_INIT_(construct, construct),
	OBS_STRUCT_INIT_(destruct, destruct),
	OBS_STRUCT_INIT_(get, get),
	OBS_STRUCT_INIT_(put, put),
};

int
main()
{
	OBS_status ret;
	OBS_slab slab;
	OBS_pa_inspector_file insp;
	void *ptr;

	obs_pa_set_current_ops(
		obs_inspect_pa_ops_file(
			&insp, obs_pa_get_current_ops(), stdout, "  pa> "));

	ret = obs_slab_init(&slab, 3000, OBS_SLAB_F_FORCE_LARGE_POLICY, OBS_ALIGN_4, &hooks, NULL);
	if (ret)
		notice("obs_slab_init", ret, 1);

	printf("slab size=%u, wastage=%u\n",
	       (unsigned)obs_slab_get_slab_size(&slab),
	       (unsigned)obs_slab_get_wastage_per_slab(&slab));

	ret = obs_slab_get(&slab, &ptr);
	if (ret)
		notice("obs_slab_get", ret, 1);
	printf("get: %p\n", ptr);
	obs_slab_put(&slab, ptr);
	printf("put(%p)\n", ptr);
	ret = obs_slab_get(&slab, &ptr);
	if (ret)
		notice("obs_slab_get", ret, 1);
	printf("get: %p\n", ptr);
	ret = obs_slab_get(&slab, &ptr);
	if (ret)
		notice("obs_slab_get", ret, 1);
	printf("get: %p\n", ptr);
	ret = obs_slab_get(&slab, &ptr);
	if (ret)
		notice("obs_slab_get", ret, 1);
	printf("get: %p\n", ptr);
	obs_slab_put(&slab, ptr);
	printf("put(%p)\n", ptr);
	obs_slab_put_all(&slab);
	printf("put_all\n");
	obs_slab_run_gc(&slab, 0);
	printf("run_gc(0)\n");
	ret = obs_slab_get(&slab, &ptr);
	if (ret)
		notice("obs_slab_get", ret, 1);
	printf("get: %p\n", ptr);

	obs_slab_uninit(&slab);
	printf("uninit\n");
	obs_pa_set_current_ops(obs_pa_get_default_ops());

	return 0;
}
