/*-
 * Copyright (c)2007 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <obs/obs.h>
#include <obs/obs_pa_insp_file.h>
#include <stdio.h>
#include <stdlib.h>

void
notice(const char *func, OBS_status errcode, int isexit)
{
	printf("%s failed: %s\n", func, obs_strstatus(errcode));
	if (isexit)
		exit(1);
}

void
dump_reg(const char *mes, const OBS_region *reg)
{
	printf("%s: ", mes);
	if (obs_region_is_valid(reg))
		printf("(%p,%u)\n", reg->head, (unsigned)reg->size);
	else
		printf("invalid\n");
}

int
main()
{
	OBS_status ret;
	OBS_stack obs;
	OBS_pa_inspector_file insp;
	OBS_region reg1;

	obs_pa_set_current_ops(
		obs_inspect_pa_ops_file(
			&insp, obs_pa_get_current_ops(), stdout, "  pa> "));

	ret = obs_stack_init(&obs, 1, 0, OBS_ALIGN_4);
	if (ret)
		notice("obs_stack_init", ret, 1);

	ret = obs_stack_alloc(&obs, 10);
	if (ret)
		notice("obs_stack_alloc", ret, 1);
	reg1 = *obs_stack_get_top(&obs);
	dump_reg("alloc(10)", obs_stack_get_top(&obs));
	ret = obs_stack_alloc(&obs, 20);
	if (ret)
		notice("obs_stack_alloc", ret, 1);
	dump_reg("alloc(20)", obs_stack_get_top(&obs));
	ret = obs_stack_truncate_after(&obs, reg1.head);
	if (ret)
		notice("obs_stack_truncate_after", ret, 1);
	dump_reg("trunc", obs_stack_get_top(&obs));
	ret = obs_stack_alloc(&obs, 4000);
	if (ret)
		notice("obs_stack_alloc", ret, 1);
	dump_reg("alloc(4000)", obs_stack_get_top(&obs));
	ret = obs_stack_resize(&obs, 8000);
	if (ret)
		notice("obs_stack_resize", ret, 1);
	dump_reg("resize(8000)", obs_stack_get_top(&obs));
	ret = obs_stack_resize(&obs, 12000);
	if (ret)
		notice("obs_stack_resize", ret, 1);
	dump_reg("resize(12000)", obs_stack_get_top(&obs));
	ret = obs_stack_resize(&obs, 16000);
	if (ret)
		notice("obs_stack_resize", ret, 1);
	dump_reg("resize(16000)", obs_stack_get_top(&obs));
	reg1 = *obs_stack_get_top(&obs);
	ret = obs_stack_alloc(&obs, 16000);
	if (ret)
		notice("obs_stack_alloc", ret, 1);
	dump_reg("alloc(16000)", obs_stack_get_top(&obs));

	ret = obs_stack_truncate_before(
		&obs, (void *)((char *)reg1.head+16000));
	if (ret)
		notice("obs_stack_truncate_before", ret, 1);
	dump_reg("trunc_before", obs_stack_get_top(&obs));

	obs_stack_uninit(&obs);
	obs_pa_set_current_ops(obs_pa_get_default_ops());

	return 0;
}
